import torch
import torch.nn as nn
import torch.nn.functional as F

def l2_normalize(x, axis=None, eps=1e-12):
    """Normalizes along dimension `axis` using an L2 norm."""
    norm = torch.sqrt(torch.sum(x * x, dim=axis, keepdim=True) + eps)
    return x / norm


def get_norm_layer(train=True, norm_type='BN', num_groups=32):
    """Normalization layer selection."""
    if norm_type == 'BN':
        return nn.BatchNorm2d if train else lambda num_features: nn.BatchNorm2d(num_features, affine=False)
    elif norm_type == 'LN':
        return nn.LayerNorm
    elif norm_type == 'GN':
        return lambda num_features: nn.GroupNorm(num_groups, num_features)
    else:
        raise NotImplementedError(f"Normalization type {norm_type} not implemented.")


def tensorflow_style_avg_pooling(x, kernel_size, stride, padding='same'):
    """TensorFlow-style average pooling in PyTorch."""
    if padding == 'same':
        pad_h = (kernel_size[0] - 1) // 2
        pad_w = (kernel_size[1] - 1) // 2
        x = F.pad(x, (pad_w, pad_w, pad_h, pad_h), mode='replicate')
    return F.avg_pool2d(x, kernel_size, stride=stride, padding=0)


def upsample(x, factor=2):
    """Upsamples tensor by a given factor using nearest neighbor interpolation."""
    return F.interpolate(x, scale_factor=factor, mode='nearest')


def dsample(x):
    """Downsamples tensor using average pooling."""
    return tensorflow_style_avg_pooling(x, (2, 2), stride=(2, 2), padding='same')



import torch 
import torch.nn as nn
from medlat.registry import register_model
from .transformer import MaskGIT
from functools import partial


@register_model("maskgit.b")
def MaskGIT_B(**kwargs):
    model = MaskGIT(
        embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model

@register_model("maskgit.l")
def MaskGIT_L(**kwargs):
    model = MaskGIT(
        embed_dim=1024, depth=16, num_heads=16, mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model

@register_model("maskgit.h")
def MaskGIT_H(**kwargs):
    model = MaskGIT(
        embed_dim=1280, depth=20, num_heads=16, mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


import torch
import torch.nn as nn
import numpy as np
from medlat.utils import init_from_ckpt
from medlat.modules.alignments import AlignmentModule
from medlat.first_stage.continuous.modules.ldm_modules import get_conv_layer
from medlat.first_stage.modules.gaussian_dist import DiagonalGaussianDistribution, _DeterministicPosterior

__all__ = [
    "AutoencoderKL",
    "AutoencoderKLTransformer",
]


class AutoencoderKL(nn.Module):
    def __init__(self,
                 encoder: nn.Module,
                 decoder: nn.Module,
                 alignment: AlignmentModule = None,
                 embed_dim: int = None,
                 kl_weight: float = 1e-6,
                 pre_post_layer: str = "conv",
                 double_z: bool = True,
                 channel_dim: int = 1,      # B, C, H, W --> 1 | B, N, C --> 2
                 ckpt_path: str = None):
        super().__init__()
        
        if pre_post_layer == "conv":
            self.dims = getattr(encoder, "dims", 2)
            conv_layer = get_conv_layer(self.dims)

        self.encoder = encoder
        self.decoder = decoder
        self.alignment = alignment
        self.double_z = double_z
        self.channel_dim = channel_dim

        self.encoder_z_channels = getattr(encoder, "z_channels", None)
        if self.encoder_z_channels is None:
            raise ValueError(f"Encoder {encoder.__class__.__name__} must define z_channels.")

        self.vae_stride = getattr(encoder, "vae_stride", None)
        if self.vae_stride is None:
            raise ValueError(f"Encoder {encoder.__class__.__name__} must define vae_stride.")

        self.kl_weight = kl_weight
        if embed_dim is None:
            embed_dim = self.encoder_z_channels
        self.embed_dim = embed_dim

        if pre_post_layer == "conv":
            self.quant_conv = conv_layer(2 * self.encoder_z_channels, 2 * embed_dim, 1)
            self.post_quant_conv = conv_layer(embed_dim, self.encoder_z_channels, 1)
        elif pre_post_layer == "linear":
            self.quant_conv = nn.Linear(2 * self.encoder_z_channels, 2 * embed_dim) 
            self.post_quant_conv = nn.Linear(embed_dim, self.encoder_z_channels)
        elif pre_post_layer == "none":
            self.quant_conv = nn.Identity()
            self.post_quant_conv = nn.Identity()
        else:
            raise ValueError(f"Invalid pre_post_layer: {pre_post_layer}")

        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)


    def get_posterior(self, x):
        h = self.encoder(x)
        moments = self.quant_conv(h)
        if not self.double_z:
            # Encoder outputs z directly (e.g. pretrained DCAE); no Gaussian
            return _DeterministicPosterior(moments)
        posterior = DiagonalGaussianDistribution(moments, channel_dim=self.channel_dim)
        return posterior

    def encode(self, x):
        h = self.encoder(x)
        moments = self.quant_conv(h)
        posterior = DiagonalGaussianDistribution(moments, channel_dim=self.channel_dim)
        return posterior.sample(), self.p_loss(posterior, x.device), None


    def decode(self, z):
        z = self.post_quant_conv(z)
        dec = self.decoder(z)
        return dec


    def p_loss(self, posterior, device): 
        kl_loss = torch.zeros((), device=device)
        if posterior is not None:
            # assume extra_result_dict contains posteriors with kl method
            if hasattr(posterior, "kl"):
                kl_loss = posterior.kl()
                kl_loss = torch.sum(kl_loss) / kl_loss.shape[0]
        kl_loss = self.kl_weight * kl_loss
        return kl_loss

    def forward(self, input, sample_posterior=True):
        posterior = self.get_posterior(input)
        if self.double_z:
            p_loss = self.p_loss(posterior, input.device)
        else:
            p_loss = torch.zeros((), device=input.device)
        if sample_posterior:
            z = posterior.sample()
        else:
            z = posterior.mode()
        dec = self.decode(z)
        
        if self.alignment is not None:
            alignment_loss, _ = self.alignment(z, input)
            return dec, p_loss + alignment_loss
        
        return dec, p_loss



class AutoencoderKLTransformer(nn.Module):
    def __init__(self,
                 encoder: nn.Module,
                 decoder: nn.Module,
                 alignment: AlignmentModule = None,
                 embed_dim: int = None,
                 kl_weight: float = 1e-6,
                 pre_post_layer: str = "linear",
                 double_z: bool = True,
                 channel_dim: int = 2,      # B, C, H, W --> 1 | B, N, C --> 2
                 ckpt_path: str = None):
        super().__init__()
        
        self.encoder = encoder
        self.decoder = decoder
        self.alignment = alignment
        self.double_z = double_z
        self.channel_dim = channel_dim

        self.encoder_z_channels = getattr(encoder, "z_channels", None)
        if self.encoder_z_channels is None:
            raise ValueError(f"Encoder {encoder.__class__.__name__} must define z_channels.")

        self.vae_stride = getattr(encoder, "vae_stride", None)
        if self.vae_stride is None:
            raise ValueError(f"Encoder {encoder.__class__.__name__} must define vae_stride.")

        self.kl_weight = kl_weight
        if embed_dim is None:
            embed_dim = self.encoder_z_channels
        self.embed_dim = embed_dim

        if pre_post_layer == "linear":
            self.quant_conv = nn.Linear(2 * self.encoder_z_channels, 2 * embed_dim) 
            self.post_quant_conv = nn.Linear(embed_dim, self.encoder_z_channels)
        elif pre_post_layer == "none":
            self.quant_conv = nn.Identity()
            self.post_quant_conv = nn.Identity()
        else:
            raise ValueError(f"Invalid pre_post_layer: {pre_post_layer}")

        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)


    def get_posterior(self, x):
        h, aux = self.encoder(x)
        moments = self.quant_conv(h)
        if not self.double_z:
            # Encoder outputs z directly (e.g. pretrained DCAE); no Gaussian
            return _DeterministicPosterior(moments), aux
        posterior = DiagonalGaussianDistribution(moments, channel_dim=self.channel_dim)
        return posterior, aux

    def encode(self, x):
        h, aux = self.encoder(x)
        moments = self.quant_conv(h)
        posterior = DiagonalGaussianDistribution(moments, channel_dim=self.channel_dim)
        return posterior.sample(), self.p_loss(posterior, x.device), None, aux


    def decode(self, z, aux=None):
        z = self.post_quant_conv(z)
        dec = self.decoder(z, ids_restore=aux["ids_restore"] if aux is not None else None)
        return dec


    def p_loss(self, posterior, device): 
        kl_loss = torch.zeros((), device=device)
        if posterior is not None:
            # assume extra_result_dict contains posteriors with kl method
            if hasattr(posterior, "kl"):
                kl_loss = posterior.kl()
                kl_loss = torch.sum(kl_loss) / kl_loss.shape[0]
        kl_loss = self.kl_weight * kl_loss
        return kl_loss

    def forward(self, input, sample_posterior=True):
        posterior, aux = self.get_posterior(input)
        if self.double_z:
            p_loss = self.p_loss(posterior, input.device)
        else:
            p_loss = torch.zeros((), device=input.device)
        if sample_posterior:
            z = posterior.sample()
        else:
            z = posterior.mode()

        print(f"z: {z.shape}")
        dec = self.decode(z, aux=aux)
        
        if self.alignment is not None:
            alignment_loss, _ = self.alignment(z, input)
            return dec, p_loss + alignment_loss
        
        return dec, p_loss







# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# GLIDE: https://github.com/openai/glide-text2im
# MAE: https://github.com/facebookresearch/mae/blob/main/models_mae.py
# --------------------------------------------------------

import torch
import torch.nn as nn
import numpy as np
import math
from einops import rearrange
from einops.layers.torch import Rearrange
from timm.models.vision_transformer import Attention, Mlp
from medlat.modules.in_and_out import PatchEmbed, ToPixel
from medlat.modules.pos_embed import get_2d_sincos_pos_embed, get_3d_sincos_pos_embed

__all__ = ["DiT"]

####################################################################



class SineLayer(nn.Module):
    """
    Paper: Implicit Neural Representation with Periodic Activ ation Function (SIREN)
    """

    def __init__(self, in_features, out_features, bias=True, is_first=False, omega_0=30):
        super().__init__()
        self.omega_0 = omega_0
        self.is_first = is_first

        self.in_features = in_features
        self.linear = nn.Linear(in_features, out_features, bias=bias)

        self.init_weights()

    def init_weights(self):
        with torch.no_grad():
            if self.is_first:
                self.linear.weight.uniform_(-1 / self.in_features, 1 / self.in_features)
            else:
                self.linear.weight.uniform_(-np.sqrt(6 / self.in_features) / self.omega_0,
                                            np.sqrt(6 / self.in_features) / self.omega_0)

    def forward(self, input):
        return torch.sin(self.omega_0 * self.linear(input))


#####################################################################

def modulate(x, shift, scale):
    return x * (1 + scale.unsqueeze(1)) + shift.unsqueeze(1)


#################################################################################
#               Embedding Layers for Timesteps and Class Labels                 #
#################################################################################

class TimestepEmbedder(nn.Module):
    """
    Embeds scalar timesteps into vector representations.
    """
    def __init__(self, hidden_size, frequency_embedding_size=256):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, hidden_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size

    @staticmethod
    def timestep_embedding(t, dim, max_period=10000):
        """
        Create sinusoidal timestep embeddings.
        :param t: a 1-D Tensor of N indices, one per batch element.
                          These may be fractional.
        :param dim: the dimension of the output.
        :param max_period: controls the minimum frequency of the embeddings.
        :return: an (N, D) Tensor of positional embeddings.
        """
        # https://github.com/openai/glide-text2im/blob/main/glide_text2im/nn.py
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=t.device)
        args = t[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
        return embedding

    def forward(self, t):
        t_freq = self.timestep_embedding(t, self.frequency_embedding_size)
        t_emb = self.mlp(t_freq)
        return t_emb


class LabelEmbedder(nn.Module):
    """
    Embeds class labels into vector representations. Also handles label dropout for classifier-free guidance.
    """
    def __init__(self, num_classes, hidden_size, dropout_prob):
        super().__init__()
        use_cfg_embedding = dropout_prob > 0
        self.embedding_table = nn.Embedding(num_classes + use_cfg_embedding, hidden_size)
        self.num_classes = num_classes
        self.dropout_prob = dropout_prob

    def token_drop(self, labels, force_drop_ids=None):
        """
        Drops labels to enable classifier-free guidance.
        """
        if force_drop_ids is None:
            drop_ids = torch.rand(labels.shape[0], device=labels.device) < self.dropout_prob
        else:
            drop_ids = force_drop_ids == 1
        labels = torch.where(drop_ids, self.num_classes, labels)
        return labels

    def forward(self, labels, train, force_drop_ids=None):
        use_dropout = self.dropout_prob > 0
        if (train and use_dropout) or (force_drop_ids is not None):
            labels = self.token_drop(labels, force_drop_ids)
        embeddings = self.embedding_table(labels)
        return embeddings


class DatasetEmbedder(nn.Module):
    """
    Embeds dataset IDs into vector representations. Also handles label dropout for classifier-free guidance.
    """
    def __init__(self, num_datasets, hidden_size, dropout_prob):
        super().__init__()
        use_cfg_embedding = dropout_prob > 0
        self.embedding_table = nn.Embedding(num_datasets + use_cfg_embedding, hidden_size)
        self.num_datasets = num_datasets
        self.dropout_prob = dropout_prob

    def token_drop(self, dataset_ids, force_drop_ids=None):
        """
        Drops dataset IDs to enable classifier-free guidance.
        """
        if force_drop_ids is None:
            drop_ids = torch.rand(dataset_ids.shape[0], device=dataset_ids.device) < self.dropout_prob
        else:
            drop_ids = force_drop_ids == 1
        dataset_ids = torch.where(drop_ids, self.num_datasets, dataset_ids)
        return dataset_ids

    def forward(self, dataset_ids, train, force_drop_ids=None):
        use_dropout = self.dropout_prob > 0
        if (train and use_dropout) or (force_drop_ids is not None):
            dataset_ids = self.token_drop(dataset_ids, force_drop_ids)
        embeddings = self.embedding_table(dataset_ids)
        return embeddings


#################################################################################
#                                 Core DiT Model                                #
#################################################################################

class DiTBlock(nn.Module):
    def __init__(self, hidden_size, num_heads, mlp_ratio=4.0, cond_dim=None, **block_kwargs):
        super().__init__()
        self.norm1 = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        self.attn = Attention(hidden_size, num_heads=num_heads, qkv_bias=True, **block_kwargs)
        self.norm2 = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        mlp_hidden_dim = int(hidden_size * mlp_ratio)
        approx_gelu = lambda: nn.GELU(approximate="tanh")
        self.mlp = Mlp(in_features=hidden_size, hidden_features=mlp_hidden_dim, act_layer=approx_gelu, drop=0)
        # FIXED: Accept cond_dim input (2x or 3x hidden_size)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(cond_dim or hidden_size, 6 * hidden_size, bias=True)
        )

    def forward(self, x, c):
        mod = self.adaLN_modulation(c)
        shift_msa, scale_msa, gate_msa, shift_mlp, scale_mlp, gate_mlp = mod.chunk(6, dim=1)
        x = x + gate_msa.unsqueeze(1) * self.attn(modulate(self.norm1(x), shift_msa, scale_msa))
        x = x + gate_mlp.unsqueeze(1) * self.mlp(modulate(self.norm2(x), shift_mlp, scale_mlp))
        return x

class FinalLayer(nn.Module):
    def __init__(self, hidden_size, patch_size, out_channels, cond_dim=None):
        super().__init__()
        self.norm_final = nn.LayerNorm(hidden_size, elementwise_affine=False, eps=1e-6)
        self.linear = nn.Linear(hidden_size, math.prod(patch_size) * out_channels, bias=True)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(cond_dim or hidden_size, 2 * hidden_size, bias=True)
        )

    def forward(self, x, c):
        shift, scale = self.adaLN_modulation(c).chunk(2, dim=1)
        x = modulate(self.norm_final(x), shift, scale)
        x = self.linear(x)
        return x



class DiT(nn.Module):
    """
    Improved Diffusion Transformer with strong dataset+label conditioning.
    """
    def __init__(
        self,
        img_size: int | tuple[int, ...] = 256,
        vae_stride: int | tuple[int, ...] = 16,
        patch_size: int | tuple[int, ...] = 2,
        in_channels=4,
        hidden_size=1152,
        depth=28,
        num_heads=16,
        mlp_ratio=4.0,
        class_label_drop_prob=0.1,  
        dataset_label_drop_prob=0.05, 
        num_classes=1000,
        dataset_num=None,
        learn_sigma=True,
        dims=2,
    ):
        super().__init__()
        self.dims = dims
        self.img_size = img_size
        self.patch_size = patch_size
        if isinstance(img_size, int):
            self.img_size = (img_size,) * self.dims
        if isinstance(patch_size, int):
            self.patch_size = (patch_size,) * self.dims

        print(f"img_size: {self.img_size}, vae_stride: {vae_stride}")
        self.img_size = tuple(i // v for i, v in zip(self.img_size, (vae_stride,) * self.dims))
        print(f"img_size: {self.img_size}, patch_size: {self.patch_size}")

        self.learn_sigma = learn_sigma
        self.in_channels = in_channels
        self.out_channels = in_channels * 2 if learn_sigma else in_channels
        self.num_heads = num_heads
        self.hidden_size = hidden_size
        
        # FIXED: Support 3 conditions (t, y, dataset) -> 3*hidden_size input
        self.cond_dim = 3 * hidden_size if dataset_num is not None else 2 * hidden_size

        self.x_embedder = PatchEmbed(to_embed='conv', img_size=self.img_size, patch_size=self.patch_size, 
                                   in_chans=in_channels, embed_dim=hidden_size)
        self.to_pixel = ToPixel(to_pixel='none', img_size=self.img_size, out_channels=self.out_channels, 
                              in_dim=hidden_size, patch_size=self.patch_size)
        self.t_embedder = TimestepEmbedder(hidden_size)
        self.y_embedder = LabelEmbedder(num_classes, hidden_size, class_label_drop_prob)
        
        self.use_dataset_conditioning = dataset_num is not None
        if self.use_dataset_conditioning:
            self.dataset_embedder = DatasetEmbedder(dataset_num, hidden_size, dataset_label_drop_prob)
        
        num_patches = self.x_embedder.num_patches
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches, hidden_size), requires_grad=False)

        self.blocks = nn.ModuleList([
            DiTBlock(hidden_size, num_heads, mlp_ratio=mlp_ratio, cond_dim=self.cond_dim) 
            for _ in range(depth)
        ])
        self.final_layer = FinalLayer(hidden_size, self.patch_size, self.out_channels, cond_dim=self.cond_dim)
        self.initialize_weights()

    def initialize_weights(self):
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)
        self.apply(_basic_init)

        if self.dims == 2:
            pos_embed = get_2d_sincos_pos_embed(self.pos_embed.shape[-1], self.x_embedder.grid_size)
        elif self.dims == 3:
            pos_embed = get_3d_sincos_pos_embed(self.pos_embed.shape[-1], self.x_embedder.grid_size)
        else:
            raise ValueError(f"dims must be 2 or 3, got {self.dims}")
        self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

        w = self.x_embedder.proj.weight.data
        nn.init.xavier_uniform_(w.view([w.shape[0], -1]))
        nn.init.constant_(self.x_embedder.proj.bias, 0)

        nn.init.normal_(self.y_embedder.embedding_table.weight, std=0.02)
        if self.use_dataset_conditioning:
            nn.init.normal_(self.dataset_embedder.embedding_table.weight, std=0.05)  # Stronger dataset signal
            
        nn.init.normal_(self.t_embedder.mlp[0].weight, std=0.02)
        nn.init.normal_(self.t_embedder.mlp[2].weight, std=0.02)

        # Zero-out adaLN modulation layers (now takes cond_dim input)
        for block in self.blocks:
            nn.init.constant_(block.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(block.adaLN_modulation[-1].bias, 0)
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].weight, 0)
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].bias, 0)
        nn.init.constant_(self.final_layer.linear.weight, 0)
        nn.init.constant_(self.final_layer.linear.bias, 0)

    def forward(self, x, t, y, dataset_id=None):
        x = self.x_embedder(x) + self.pos_embed
        t_emb = self.t_embedder(t)
        y_emb = self.y_embedder(y, self.training)
        
        # FIXED: Concatenate instead of add to avoid signal interference
        c_list = [t_emb, y_emb]
        if self.use_dataset_conditioning and dataset_id is not None:
            ds_emb = self.dataset_embedder(dataset_id, self.training)
            c_list.append(ds_emb)
        
        c = torch.cat(c_list, dim=1)  # (N, cond_dim)
        
        for block in self.blocks:
            x = block(x, c)
        x = self.final_layer(x, c)
        x = self.to_pixel(x)
        return x

    def forward_with_cfg(self, x, t, y, cfg_scale, dataset_id=None, cfg_dataset=True):
        """
        Flexible CFG: control whether dataset_id gets CFG or stays fixed.
        
        Args:
            cfg_dataset (bool): If True, apply CFG to both labels + dataset_id
                            If False, apply CFG only to labels (dataset_id fixed)
                            
        Perfect for:
        - cfg_dataset=True:  Full control (dataset switching)
        - cfg_dataset=False: Fixed dataset + label control only (RetinaMNIST classes)
        """
        batch_size = x.shape[0]
        
        # DOUBLE x and t
        x_doubled = torch.cat([x, x], dim=0)
        t_doubled = torch.cat([t, t], dim=0)
        
        # CLASS-LABEL CFG (always applied)
        y_cond = y
        y_uncond = torch.full_like(y, self.y_embedder.num_classes)
        y_doubled = torch.cat([y_cond, y_uncond], dim=0)
        
        # DATASET CFG (user controlled)
        if self.use_dataset_conditioning and dataset_id is not None:
            if cfg_dataset:
                # Full CFG: conditional + unconditional dataset
                ds_cond = dataset_id
                ds_uncond = torch.full_like(dataset_id, self.dataset_embedder.num_datasets)
                dataset_id_doubled = torch.cat([ds_cond, ds_uncond], dim=0)
            else:
                # FIXED dataset: same for both passes
                dataset_id_doubled = torch.cat([dataset_id, dataset_id], dim=0)
        else:
            dataset_id_doubled = None
        
        # Forward pass
        model_out_doubled = self.forward(x_doubled, t_doubled, y_doubled, dataset_id_doubled)
        
        # Apply CFG
        cond_out, uncond_out = torch.split(model_out_doubled, batch_size, dim=0)
        guided_out = uncond_out + cfg_scale * (cond_out - uncond_out)
        
        return guided_out[:batch_size]




if __name__ == "__main__":
    model = DiT(img_size=[32, 32, 32], patch_size=[2, 2, 2])
    x = torch.randn(1, 4, 32, 32, 32) # (N, C, H, W, D)
    t = torch.randint(0, 1000, (1,))
    y = torch.randint(0)
    out = model(x, t, y)
    print(out.shape)


import torch
from medlat.registry import register_model
from .titok.titok import TiTok
from .maetok.modules.vit_models import MAETokViTEncoder, MAETokViTDecoder
from .maetok.maetok import MaskAEModel

from typing import Optional
from medlat.first_stage.discrete.vq_models import VQModel
from medlat.first_stage.discrete.quantizer.quantize import SoftVectorQuantizer
from medlat.modules.alignments import *


@register_model("token.titok.s_128")
def TiTok_S_128(
    img_size: int | tuple[int, ...],
    patch_size: int | tuple[int, ...] = 16, 
    hidden_size=512,
    in_channels=3,
    out_channels=3,
    depth=8,
    num_heads=8,
    num_latent_tokens=128,
    token_size=12,
    codebook_size=4096,
    quantizer_loss_weight=1.0,
    pixel_vqgan: Optional[VQModel] = None, # VQModel = PretrainedTokenizer("/vol/miltank/users/bubeckn/1d-tokenizer/maskgit-vqgan-imagenet-f16-256.bin")
    stage="1", 
    quantize_mode="vq",
    **kwargs,
):
    return TiTok(
        img_size=img_size,
        patch_size=patch_size,
        hidden_size=hidden_size,
        in_channels=in_channels,
        out_channels=out_channels,
        depth=depth,
        num_heads=num_heads,
        num_latent_tokens=num_latent_tokens,
        token_size=token_size,
        codebook_size=codebook_size,
        quantizer_loss_weight=quantizer_loss_weight,
        pixel_vqgan=pixel_vqgan,
        stage=stage,
        quantize_mode=quantize_mode,
        **kwargs,
    )

@register_model("token.titok.s_128_e2e")
def TiTok_S_128_E2E(**kwargs):
    return TiTok_S_128(num_latent_tokens=128, patch_size=8, stage="e2e", **kwargs)

@register_model("token.titok.s_256_e2e")
def TiTok_S_256_E2E(**kwargs):
    return TiTok_S_128(num_latent_tokens=256, patch_size=8, stage="e2e", **kwargs)

@register_model("token.titok.s_512_e2e")
def TiTok_S_512_E2E(**kwargs):
    return TiTok_S_128(num_latent_tokens=512, patch_size=8, stage="e2e", **kwargs)


@register_model("token.titok.b_64")
def TiTok_B_64(img_size: int | tuple[int, ...],
                patch_size: int | tuple[int, ...] = 16, 
                hidden_size=768,
                in_channels=3,
                out_channels=3,
                depth=12,
                num_heads=12,
                num_latent_tokens=64,
                token_size=12,
                codebook_size=4096,
                quantizer_loss_weight=1.0,
                pixel_vqgan: Optional[VQModel] = None, # VQModel = PretrainedTokenizer("/vol/miltank/users/bubeckn/1d-tokenizer/maskgit-vqgan-imagenet-f16-256.bin"), 
                stage="1", 
                quantize_mode="vq",
                **kwargs):
    return TiTok(img_size=img_size, patch_size=patch_size, hidden_size=hidden_size, in_channels=in_channels, out_channels=out_channels, depth=depth, num_heads=num_heads, num_latent_tokens=num_latent_tokens, token_size=token_size, codebook_size=codebook_size, quantizer_loss_weight=quantizer_loss_weight, pixel_vqgan=pixel_vqgan, stage=stage, quantize_mode=quantize_mode, **kwargs)

@register_model("token.titok.b_128_p8_e2e")
def TiTok_B_128_p8_E2E(**kwargs):
    return TiTok_B_64(num_latent_tokens=128, patch_size=8, stage="e2e", **kwargs)

@register_model("token.titok.b_256_p8_e2e")
def TiTok_B_256_p8_E2E(**kwargs):
    return TiTok_B_64(num_latent_tokens=256, patch_size=8, stage="e2e", **kwargs)

@register_model("token.titok.b_512_p8_e2e")
def TiTok_B_512_p8_E2E(**kwargs):
    return TiTok_B_64(num_latent_tokens=512, patch_size=8, stage="e2e", **kwargs)

@register_model("token.titok.l_32")
def TiTok_L_32(img_size: int | tuple[int, ...],
                patch_size: int | tuple[int, ...] = 16, 
                hidden_size=1024,
                in_channels=3,
                out_channels=3,
                depth=24,
                num_heads=16,
                num_latent_tokens=32,
                token_size=12,
                codebook_size=4096,
                quantizer_loss_weight=1.0,
                pixel_vqgan: Optional[VQModel] = None, # VQModel = PretrainedTokenizer("/vol/miltank/users/bubeckn/1d-tokenizer/maskgit-vqgan-imagenet-f16-256.bin"), 
                stage="1", 
                quantize_mode="vq",
                **kwargs):
    return TiTok(img_size=img_size, patch_size=patch_size, hidden_size=hidden_size, in_channels=in_channels, out_channels=out_channels, depth=depth, num_heads=num_heads, num_latent_tokens=num_latent_tokens, token_size=token_size, codebook_size=codebook_size, quantizer_loss_weight=quantizer_loss_weight, pixel_vqgan=pixel_vqgan, stage=stage, quantize_mode=quantize_mode, **kwargs)

@register_model("token.titok.l_64_e2e")
def TiTok_L_64_E2E(**kwargs):
    return TiTok_L_32(num_latent_tokens=64, stage="e2e", **kwargs)

@register_model("token.titok.l_128_e2e")
def TiTok_L_128_E2E(**kwargs):
    return TiTok_L_32(num_latent_tokens=128, stage="e2e", **kwargs)

@register_model("token.titok.l_256_e2e")
def TiTok_L_256_E2E(**kwargs):
    return TiTok_L_32(num_latent_tokens=256, stage="e2e", **kwargs)

@register_model("token.titok.l_512_e2e")
def TiTok_L_512_E2E(**kwargs):
    return TiTok_L_32(num_latent_tokens=512, stage="e2e", **kwargs)


###### MAETOK #######
@register_model("token.maetok.s_128")
def MAETok_S_128(img_size: int = 256, 
                 patch_size: int = 16,
                 codebook_embed_dim: int = 32,
                 num_latent_tokens: int = 128,
                 enc_embed_dim: int = 512,
                 dec_embed_dim: int = 512,
                 enc_depth: int = 8,
                 dec_depth: int = 8,
                 enc_num_heads: int = 8,
                 dec_num_heads: int = 8,
                 aux_hog_dec: bool = True,
                 aux_dino_dec: bool = True,
                 aux_clip_dec: bool = True,
                 aux_biomed_clip_dec: bool = False, **kwargs):  
    return MaskAEModel(img_size=img_size, base_img_size=img_size, patch_size=patch_size, codebook_embed_dim=codebook_embed_dim, num_latent_tokens=num_latent_tokens, enc_embed_dim=enc_embed_dim, dec_embed_dim=dec_embed_dim, enc_depth=enc_depth, dec_depth=dec_depth, enc_num_heads=enc_num_heads, dec_num_heads=dec_num_heads, aux_hog_dec=aux_hog_dec, aux_dino_dec=aux_dino_dec, aux_clip_dec=aux_clip_dec, aux_biomed_clip_dec=aux_biomed_clip_dec, **kwargs)


@register_model("token.maetok.s_256")
def MAETok_S_256(**kwargs):
    return MAETok_S_128(num_latent_tokens=256, **kwargs)

@register_model("token.maetok.s_512")
def MAETok_S_512(**kwargs):
    return MAETok_S_128(num_latent_tokens=512, **kwargs)


@register_model("token.maetok.b_128_p16",
                code_url="https://github.com/Hhhhhhao/continuous_tokenizer",
                paper_url="https://arxiv.org/pdf/2502.03444",)
def MAETok_B_128(img_size: int = 256, 
                 patch_size: int = 16,
                 codebook_embed_dim: int = 32,
                 num_latent_tokens: int = 128,
                 aux_hog_dec: bool = True,
                 aux_dino_dec: bool = True,
                 aux_clip_dec: bool = True,
                 aux_biomed_clip_dec: bool = False,
                 **kwargs):
    return MaskAEModel(img_size=img_size, base_img_size=img_size, patch_size=patch_size, codebook_embed_dim=codebook_embed_dim, num_latent_tokens=num_latent_tokens, aux_hog_dec=aux_hog_dec, aux_dino_dec=aux_dino_dec, aux_clip_dec=aux_clip_dec, aux_biomed_clip_dec=aux_biomed_clip_dec, **kwargs)

@register_model("token.maetok.b_128_p8")
def MAETok_B_128_p8(**kwargs):
    return MAETok_B_128(num_latent_tokens=128, patch_size=8, **kwargs)

@register_model("token.maetok.b_256_p8")
def MAETok_B_256_p8(**kwargs):
    return MAETok_B_128(num_latent_tokens=256, patch_size=8, **kwargs)

@register_model("token.maetok.b_512_p8")
def MAETok_B_512_p8(**kwargs):
    return MAETok_B_128(num_latent_tokens=512, patch_size=8, **kwargs)


@register_model("token.maetok.l_128")
def MAETok_L_128(img_size: int = 256, 
                 patch_size: int = 16,
                 codebook_embed_dim: int = 32,
                 num_latent_tokens: int = 128,
                 enc_embed_dim: int = 1024,
                 dec_embed_dim: int = 1024,
                 enc_depth: int = 24,
                 dec_depth: int = 24,
                 enc_num_heads: int = 16,
                 dec_num_heads: int = 16,
                 aux_hog_dec: bool = True,
                 aux_dino_dec: bool = True,
                 aux_clip_dec: bool = True,
                 aux_biomed_clip_dec: bool = False, **kwargs):  
    return MaskAEModel(img_size=img_size, base_img_size=img_size, patch_size=patch_size, codebook_embed_dim=codebook_embed_dim, num_latent_tokens=num_latent_tokens, enc_embed_dim=enc_embed_dim, dec_embed_dim=dec_embed_dim, enc_depth=enc_depth, dec_depth=dec_depth, enc_num_heads=enc_num_heads, dec_num_heads=dec_num_heads, aux_hog_dec=aux_hog_dec, aux_dino_dec=aux_dino_dec, aux_clip_dec=aux_clip_dec, aux_biomed_clip_dec=aux_biomed_clip_dec, **kwargs)

@register_model("token.maetok.l_256")
def MAETok_L_256(**kwargs):
    return MAETok_L_128(num_latent_tokens=256, **kwargs)

@register_model("token.maetok.l_512")
def MAETok_L_512(**kwargs):
    return MAETok_L_128(num_latent_tokens=512, **kwargs)



###### SoftVQ #######

@register_model("token.softvq.s_t32_d32")
def SoftVQ_S_T32_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 32,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)

@register_model("token.softvq.s_t64_d32")
def SoftVQ_S_T64_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 64,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)


@register_model("token.softvq.b_t32_d32")
def SoftVQ_B_T32_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 32,
    embed_dim_encoder: int = 768,
    embed_dim_decoder: int = 768,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 12,
    num_heads_decoder: int = 12,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)

@register_model("token.softvq.b_t64_d32")
def SoftVQ_B_T64_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 64,
    embed_dim_encoder: int = 768,
    embed_dim_decoder: int = 768,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 12,
    num_heads_decoder: int = 12,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)

@register_model("token.softvq.bl_t32_d32")
def SoftVQ_BL_T32_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 32,
    embed_dim_encoder: int = 768,
    embed_dim_decoder: int = 1024,
    depth_encoder: int = 12,
    depth_decoder: int = 24,
    num_heads_encoder: int = 12,
    num_heads_decoder: int = 16,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)

@register_model("token.softvq.bl_t64_d32")
def SoftVQ_BL_T64_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 64,
    embed_dim_encoder: int = 768,
    embed_dim_decoder: int = 1024,
    depth_encoder: int = 12,
    depth_decoder: int = 24,
    num_heads_encoder: int = 12,
    num_heads_decoder: int = 16,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, **kwargs)

@register_model("token.softvq.l_t32_d32")
def SoftVQ_L_T32_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 32,
    embed_dim_encoder: int = 1024,
    embed_dim_decoder: int = 1024,
    depth_encoder: int = 24,
    depth_decoder: int = 24,
    num_heads_encoder: int = 16,
    num_heads_decoder: int = 16,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )
    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )
    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)

@register_model("token.softvq.l_t64_d32")
def SoftVQ_L_T64_D32(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    num_latent_tokens: int = 64,
    embed_dim_encoder: int = 1024,
    embed_dim_decoder: int = 1024,
    depth_encoder: int = 24,
    depth_decoder: int = 24,
    num_heads_encoder: int = 16,
    num_heads_decoder: int = 16,
    mlp_ratio: float = 4.0,
    # quantizer config
    n_e: int = 8192,
    e_dim: int = 32,
    entropy_loss_weight: float = 0.01,
    entropy_loss_temperature: float = 0.01,
    entropy_gamma: float = 1.0,
    tau: float = 0.07,
    use_norm: bool = True,
    # alignment config
    foundation_type: str = "dinov2",
    **kwargs
    ):

    encoder = MAETokViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
    )
    decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
    )

    aux_dino_decoder = MAETokViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        num_latent_tokens=num_latent_tokens,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=mlp_ratio,
        codebook_embed_dim=e_dim,
        to_pixel='identity',
    )

    quantizer = SoftVectorQuantizer(
        n_e=n_e,
        e_dim=e_dim,
        entropy_loss_weight=entropy_loss_weight,
        entropy_loss_temperature=entropy_loss_temperature,
        entropy_gamma=entropy_gamma,
        tau=tau,
        use_norm=use_norm,
    )
    alignment = DinoAlignment(decoder=aux_dino_decoder, codebook_embed_dim=e_dim, img_size=img_size)
    return VQModel(encoder=encoder, decoder=decoder, quantizer=quantizer, alignment=alignment, pre_post_layer="linear", **kwargs)




"""
Integrated Vision Transformer Models for MAE-Tok

This module contains self-contained implementations of TimmViTEncoder and TimmViTDecoder
that include all the Vision Transformer logic directly, without relying on external timm models.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import math
from functools import partial
import scipy.stats as stats
from typing import Optional, Tuple
from torch import Tensor

from medlat.modules.in_and_out import PatchEmbed, ToPixel
from medlat.modules.pos_embed import (
    get_sincos_pos_embed, 
    apply_rotary_emb, 
    get_rope_tensor_2d,
    get_rope_tensor_3d
)


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_norm=False, attn_drop=0., proj_drop=0., norm_layer=nn.LayerNorm):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, attn_mask=None, rope_tensor=None, num_prefix_tokens=1, num_latent_tokens=32):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim)
        qkv = qkv.permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q, k, v = q.contiguous(), k.contiguous(), v.contiguous()
        
        q, k = self.q_norm(q), self.k_norm(k)
        
        if rope_tensor is not None:
            # Apply RoPE to the image tokens (skip prefix and latent tokens)
            # rope_tensor shape: (num_positions, head_dim * 2) where num_positions = grid_h * grid_w
            # x shape: (B, num_heads, seq_len, head_dim)
            # We need to apply RoPE to image tokens only
            if num_latent_tokens == 0:
                # Apply RoPE to all tokens except prefix
                num_img_tokens = q.shape[2] - num_prefix_tokens
                q_img = q[:, :, num_prefix_tokens:]  # (B, num_heads, num_img_tokens, head_dim)
                k_img = k[:, :, num_prefix_tokens:]
            else:
                # Apply RoPE only to image tokens (between prefix and latent)
                num_img_tokens = q.shape[2] - num_prefix_tokens - num_latent_tokens
                q_img = q[:, :, num_prefix_tokens:-num_latent_tokens]  # (B, num_heads, num_img_tokens, head_dim)
                k_img = k[:, :, num_prefix_tokens:-num_latent_tokens]
            
            # Slice rope_tensor to match number of image tokens
            # rope_tensor: (num_positions, head_dim * 2)
            # We need: (num_img_tokens, head_dim * 2)
            if rope_tensor.shape[0] < num_img_tokens:
                raise ValueError(
                    f"rope_tensor has {rope_tensor.shape[0]} positions but need {num_img_tokens} "
                    f"for image tokens. seq_len={q.shape[2]}, num_prefix_tokens={num_prefix_tokens}, "
                    f"num_latent_tokens={num_latent_tokens}"
                )
            rope_tensor_sliced = rope_tensor[:num_img_tokens]
            
            # Apply RoPE
            q_rot = apply_rotary_emb(q_img, rope_tensor_sliced)
            k_rot = apply_rotary_emb(k_img, rope_tensor_sliced)
            
            # Concatenate back
            if num_latent_tokens == 0:
                q = torch.cat([q[:, :, :num_prefix_tokens], q_rot], dim=2)
                k = torch.cat([k[:, :, :num_prefix_tokens], k_rot], dim=2)
            else:
                q = torch.cat([q[:, :, :num_prefix_tokens], q_rot, q[:, :, -num_latent_tokens:]], dim=2)
                k = torch.cat([k[:, :, :num_prefix_tokens], k_rot, k[:, :, -num_latent_tokens:]], dim=2)
        
        q = q * self.scale
        attn = q @ k.transpose(-2, -1)
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = attn @ v
        x = x.transpose(1, 2).reshape(B, N, C)
        
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

class SwiGLUFFN(nn.Module):
    """Swish-Gated Linear Unit Feed-Forward Network."""

    def __init__(self, in_features: int, hidden_features: int = None, out_features: int = None):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.w12 = nn.Linear(in_features, 2 * hidden_features)
        self.w3 = nn.Linear(hidden_features, out_features)

    def forward(self, x: Tensor) -> Tensor:
        x1, x2 = self.w12(x).chunk(2, dim=-1)
        return self.w3(F.silu(x1) * x2)

class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks)."""
    def __init__(self, drop_prob: float = 0.0):
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x: Tensor) -> Tensor:
        if self.drop_prob == 0.0 or not self.training:
            return x
        keep_prob = 1 - self.drop_prob
        random_tensor = keep_prob + torch.rand(x.shape[0], 1, 1, device=x.device, dtype=x.dtype)
        random_tensor.floor_()
        output = x.div(keep_prob) * random_tensor
        return output


class Block(nn.Module):
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_norm=False, drop=0., attn_drop=0.,
                 drop_path=0., mlp_layer=Mlp, attn_layer=Attention, act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = attn_layer(dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_norm=qk_norm, 
                             attn_drop=attn_drop, proj_drop=drop, norm_layer=norm_layer)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = mlp_layer(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, rope_tensor=None, num_prefix_tokens=1, num_latent_tokens=32):
        x = x + self.drop_path(self.attn(self.norm1(x), rope_tensor=rope_tensor, 
                                        num_prefix_tokens=num_prefix_tokens, num_latent_tokens=num_latent_tokens))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


def modulate(x, shift, scale):
    return x * scale + shift


class MoVQNorm(nn.Module):
    def __init__(self, latent_dim, dim):
        super().__init__()
        self.norm = nn.LayerNorm(dim, elementwise_affine=False, eps=1e-06)
        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(latent_dim, 2 * dim, bias=True)
        )

    def forward(self, x, interpolate_zq, num_prefix_tokens=1, num_latent_tokens=32):
        shift, scale = self.adaLN_modulation(interpolate_zq).chunk(2, dim=-1)
        norm_out = self.norm(x)
        
        norm_out = torch.cat([
            norm_out[:, :-num_latent_tokens],
            modulate(norm_out[:, -num_latent_tokens:], shift, scale)
        ], dim=1)
            
        return norm_out 


class MoVQBlockv2(nn.Module):
    def __init__(
            self,
            dim: int,
            num_heads: int,
            mlp_ratio: float = 4.,
            qkv_bias: bool = False,
            qk_norm: bool = False,
            drop: float = 0.,
            attn_drop: float = 0.,
            init_values: Optional[float] = None,
            drop_path: float = 0.,
            act_layer: nn.Module = nn.GELU,
            norm_layer: nn.Module = nn.LayerNorm,
            mlp_layer: nn.Module = Mlp,
            attn_layer: nn.Module = Attention,
            latent_dim: int = None,
    ) -> None:
        assert latent_dim is not None
        super().__init__()
        self.norm1 = norm_layer(dim, elementwise_affine=False, eps=1e-06)
        self.attn = attn_layer(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            attn_drop=attn_drop,
            proj_drop=drop,
            norm_layer=norm_layer,
        )
        # self.ls1 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        # self.ls1 = nn.Identity()
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim, elementwise_affine=False, eps=1e-06)
        self.mlp = mlp_layer(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=act_layer,
            drop=drop,
        )
        # self.ls2 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        # self.ls2 = nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(latent_dim, 4 * dim, bias=True)
        )

    def forward(self, x: torch.Tensor, interpolate_zq: torch.Tensor,
                attn_mask: torch.Tensor = None, 
                rope_tensor=None, num_prefix_tokens=1, num_latent_tokens=32) -> torch.Tensor:
        
        shift_msa, scale_msa, shift_mlp, scale_mlp = self.adaLN_modulation(interpolate_zq).chunk(4, dim=-1)
        
        # attention
        attn_out = self.norm1(x)
        
        attn_out = torch.cat([
            attn_out[:, :-num_latent_tokens],
            modulate(attn_out[:, -num_latent_tokens:], shift_msa, scale_msa)
        ], dim=1)
            
        attn_out = self.attn(attn_out, attn_mask, rope_tensor=rope_tensor, num_prefix_tokens=num_prefix_tokens, num_latent_tokens=num_latent_tokens)
        x = x + self.drop_path1(attn_out)
        
        # mlp
        mlp_out = self.norm2(x)
        

        mlp_out = torch.cat([
            mlp_out[:, :-num_latent_tokens],
            modulate(mlp_out[:, -num_latent_tokens:], shift_mlp, scale_mlp)
        ], dim=1)
            
        mlp_out = self.mlp(mlp_out)
        x = x + self.drop_path2(mlp_out)
        
        return x

class MAETokViTEncoder(nn.Module):
    """
    Integrated Vision Transformer Encoder that includes all logic directly.
    """
    def __init__(self, in_channels=3, num_latent_tokens=32,
                 embed_dim=768, depth=12, num_heads=12, mlp_ratio=4.,
                 img_size=224, patch_size=16, drop_path_rate=0.0,
                 qkv_bias=True, qk_norm=False, attn_drop=0., proj_drop=0.,
                 rope_theta=100.0, rope_mixed=False, use_rope=False, use_ape=True,
                 token_drop=0.4, token_drop_max=0.6, base_img_size=224):
        super().__init__()

        self.num_latent_tokens = num_latent_tokens
        self.img_size = img_size
        self.patch_size = patch_size
        self.embed_dim = embed_dim
        self.num_prefix_tokens = 1  # CLS token
        self.z_channels = embed_dim
        # Patch embedding
        self.patch_embed = PatchEmbed(to_embed='conv', img_size=img_size, patch_size=patch_size, 
                                     in_chans=in_channels, embed_dim=embed_dim)
        self.num_img_tokens = self.patch_embed.num_patches
        
        # Position embedding - use sinusoidal from pos_embed.py
        self.use_ape = use_ape
        if self.use_ape:
            self.pos_embed = nn.Parameter(torch.zeros(1,self.num_img_tokens + self.num_prefix_tokens, embed_dim))
            nn.init.trunc_normal_(self.pos_embed, std=0.02)
        else:
            self.pos_embed = None
        
        # CLS token
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        nn.init.trunc_normal_(self.cls_token, std=0.02)
        
        # Dropout
        self.pos_drop = nn.Dropout(p=proj_drop)
        
        # Latent tokens
        if self.num_latent_tokens:
            self.latent_tokens = nn.Parameter(torch.zeros(1, self.num_latent_tokens, embed_dim))
            nn.init.normal_(self.latent_tokens, std=0.02)
            # Latent position embeddings remain learnable
            self.latent_pos_embed = nn.Parameter(torch.zeros(1, self.num_latent_tokens, embed_dim))
            nn.init.trunc_normal_(self.latent_pos_embed, std=0.02)

        # Token dropout
        self.token_drop = token_drop > 0.0
        if self.token_drop:
            self.mask_ratio_generator = stats.truncnorm(
                (token_drop - token_drop_max) / 0.25, 0, loc=token_drop_max, scale=0.25
            )
            self.mask_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
            nn.init.normal_(self.mask_token, std=0.02)

        # RoPE
        self.use_rope = use_rope
        if self.use_rope:
            self.use_ape = False
        self.rope_mixed = rope_mixed
        self.rope_theta = rope_theta
        
        # Determine if we're using 3D based on patch_embed dimensions
        self.dims = self.patch_embed.dims
        # Force absolute positional embeddings for 3D data
        if self.dims == 3:
            self.use_rope = False
            self.use_ape = True
        
        # Initialize RoPE tensors using pos_embed.py functions
        if self.use_rope:
            # Calculate grid dimensions for RoPE
            if isinstance(img_size, int):
                grid_h = grid_w = img_size // patch_size
            elif len(img_size) == 2:
                grid_h = img_size[0] // patch_size
                grid_w = img_size[1] // patch_size
            else:
                # 3D - RoPE not supported, should have been disabled above
                grid_h = grid_w = 1
            
            # Use get_rope_tensor_2d from pos_embed.py
            # head_dim is embed_dim // num_heads
            head_dim = embed_dim // num_heads
            rope_tensor = get_rope_tensor_2d(head_dim, grid_h, grid_w)
            self.register_buffer('rope_tensor', rope_tensor)
            
            # Store grid dimensions for dynamic resizing
            self.grid_h = grid_h
            self.grid_w = grid_w
            
        # Transformer blocks
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]
        self.blocks = nn.ModuleList([
            Block(dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                  qk_norm=qk_norm, drop=proj_drop, attn_drop=attn_drop, drop_path=dpr[i])
            for i in range(depth)
        ])
        
        # Final normalization
        self.norm = nn.LayerNorm(embed_dim)

    def no_weight_decay(self):
        return {'pos_embed', 'cls_token', 'latent_tokens', 'latent_pos_embed'}

    def sample_orders(self, bsz, seq_len):
        orders = []
        for _ in range(bsz):
            order = np.array(list(range(seq_len)))
            np.random.shuffle(order)
            orders.append(order)
        orders = torch.Tensor(np.array(orders)).long()
        return orders

    def random_masking(self, x, orders):
        bsz, seq_len, embed_dim = x.shape
        mask_rate = self.mask_ratio_generator.rvs(1)[0]
        num_masked_tokens = int(np.ceil(seq_len * mask_rate))
        mask = torch.zeros(bsz, seq_len, device=x.device)
        mask = torch.scatter(mask, dim=-1, index=orders[:, :num_masked_tokens],
                           src=torch.ones(bsz, num_masked_tokens, device=x.device))
        return mask

    def forward(self, x, return_mask=False):
        if x.ndim == 4:
            B, _, H, W = x.shape
        elif x.ndim == 5:
            B, _, H, W, D = x.shape
        else:
            raise ValueError(f"Input tensor must be 4D or 5D, got {x.ndim}D")
        
        # Patch embedding
        x = self.patch_embed(x)  # (B, num_patches, embed_dim)
        # Add CLS token
        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.cat([cls_tokens, x], dim=1)  # (B, 1 + num_patches, embed_dim)
        
        # Token dropout
        if self.token_drop and self.training:
            orders = self.sample_orders(bsz=x.size(0), seq_len=x.size(1)).to(x.device)
            mask = self.random_masking(x, orders).unsqueeze(-1)
            x = torch.where(mask.bool(), self.mask_token, x)
        else:
            mask = None
        
        # Position embedding
        if self.use_ape and self.pos_embed is not None:
            x = x + self.pos_embed
        x = self.pos_drop(x)  # (B, 1 + num_patches, embed_dim)
        

        # Add latent tokens
        if self.num_latent_tokens:
            z = self.latent_tokens.expand(x.size(0), -1, -1)  # (B, num_latent_tokens, embed_dim)
            x = torch.cat([x, z + self.latent_pos_embed], dim=1)  # (B, 1 + num_patches + num_latent_tokens, embed_dim)
        # Forward through transformer blocks
        if self.use_ape:
            for blk in self.blocks:
                x = blk(x)
        elif self.use_rope:
            # Get or compute RoPE tensor for current spatial dimensions
            # patch_size is stored as int in encoder
            patch_h = self.patch_size if isinstance(self.patch_size, int) else self.patch_size[0]
            patch_w = self.patch_size if isinstance(self.patch_size, int) else self.patch_size[1] if len(self.patch_size) > 1 else self.patch_size[0]
            
            current_grid_h = H // patch_h
            current_grid_w = W // patch_w
            
            # Recompute RoPE tensor if spatial dimensions changed
            if hasattr(self, 'grid_h') and (current_grid_h != self.grid_h or current_grid_w != self.grid_w):
                head_dim = self.embed_dim // self.blocks[0].attn.num_heads
                rope_tensor = get_rope_tensor_2d(head_dim, current_grid_h, current_grid_w)
                rope_tensor = rope_tensor.to(x.device)
            else:
                rope_tensor = self.rope_tensor
            
            for blk in self.blocks:
                x = blk(x, rope_tensor=rope_tensor, num_prefix_tokens=self.num_prefix_tokens, 
                       num_latent_tokens=self.num_latent_tokens)
        else:
            for blk in self.blocks:
                x = blk(x)
        
        # Final normalization
        x = self.norm(x)
        
        # Extract output
        if self.num_latent_tokens:
            out = x[:, -self.num_latent_tokens:]  # (B, num_latent_tokens, embed_dim)
        else:
            out = x[:, self.num_prefix_tokens:]  # (B, num_patches, embed_dim)
        
        if return_mask:
            return out, mask
        else:
            return out


class MAETokViTDecoder(nn.Module):
    """
    Integrated Vision Transformer Decoder that includes all logic directly.
    """
    def __init__(self, in_channels=3, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4.,
                 img_size=224, patch_size=16, drop_path_rate=0.0,
                 qkv_bias=True, qk_norm=False, attn_drop=0., proj_drop=0.,
                 num_latent_tokens=32, to_pixel='linear', codebook_embed_dim=32,
                 rope_theta=100.0, rope_mixed=False, use_rope=False, use_ape=True,
                 cls_token=True, base_img_size=224, use_movq=False):
        super().__init__()

        # Handle img_size: can be int (2D), tuple of 2 (2D), or tuple of 3 (3D)
        if isinstance(img_size, int):
            self.img_size = (img_size, img_size)
            self.dims = 2
        elif len(img_size) == 2:
            self.img_size = img_size
            self.dims = 2
        elif len(img_size) == 3:
            self.img_size = img_size
            self.dims = 3
        else:
            raise ValueError(f"img_size must be int or tuple of length 2 or 3, got {img_size}")
        
        # Handle patch_size: can be int (2D), tuple of 2 (2D), or tuple of 3 (3D)
        if isinstance(patch_size, int):
            if self.dims == 2:
                self.patch_size = (patch_size, patch_size)
            else:  # 3D
                self.patch_size = (patch_size, patch_size, patch_size)
        elif len(patch_size) == 2:
            if self.dims == 2:
                self.patch_size = patch_size
            else:
                raise ValueError(f"patch_size tuple of length 2 not compatible with 3D img_size {img_size}")
        elif len(patch_size) == 3:
            if self.dims == 3:
                self.patch_size = patch_size
            else:
                raise ValueError(f"patch_size tuple of length 3 not compatible with 2D img_size {img_size}")
        else:
            raise ValueError(f"patch_size must be int or tuple of length 2 or 3, got {patch_size}")
        
        # Calculate number of image tokens
        if self.dims == 2:
            self.num_img_tokens = (self.img_size[0] // self.patch_size[0]) * (self.img_size[1] // self.patch_size[1])
        else:  # 3D
            self.num_img_tokens = (self.img_size[0] // self.patch_size[0]) * (self.img_size[1] // self.patch_size[1]) * (self.img_size[2] // self.patch_size[2])
        
        self.embed_dim = embed_dim
        self.num_prefix_tokens = 1 if cls_token else 0
        self.num_latent_tokens = num_latent_tokens
        self.use_movq = use_movq

        # Mask token
        self.mask_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        nn.init.normal_(self.mask_token, std=0.02)

        # Latent position embedding
        if self.num_latent_tokens:
            self.latent_pos_embed = nn.Parameter(torch.zeros(1, self.num_latent_tokens, embed_dim))
            nn.init.trunc_normal_(self.latent_pos_embed, std=0.02)

        # Position embedding - use sinusoidal from pos_embed.py
        self.use_ape = use_ape
        if self.use_ape:
            self.pos_embed = nn.Parameter(torch.zeros(1,self.num_img_tokens + self.num_prefix_tokens, embed_dim))
            nn.init.trunc_normal_(self.pos_embed, std=0.02)
        else:
            self.pos_embed = None

        # CLS token
        if cls_token:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
            nn.init.trunc_normal_(self.cls_token, std=0.02)
        else:
            self.cls_token = None

        # Dropout
        self.pos_drop = nn.Dropout(p=proj_drop)

        # To pixel head
        self.to_pixel = ToPixel(to_pixel=to_pixel, img_size=self.img_size, out_channels=in_channels,
                               in_dim=embed_dim, patch_size=self.patch_size)

        # RoPE
        self.use_rope = use_rope
        if self.use_rope:
            self.use_ape = False
        self.rope_mixed = rope_mixed
        self.rope_theta = rope_theta
        
        # Force absolute positional embeddings for 3D data
        if self.dims == 3:
            self.use_rope = False
            self.use_ape = True
        
        # Initialize RoPE tensors using pos_embed.py functions
        if self.use_rope:
            # Calculate grid dimensions for RoPE
            if isinstance(img_size, int):
                grid_h = grid_w = img_size // patch_size
            elif len(img_size) == 2:
                grid_h = img_size[0] // patch_size
                grid_w = img_size[1] // patch_size
            else:
                # 3D - RoPE not supported, should have been disabled above
                grid_h = grid_w = 1
            
            # Use get_rope_tensor_2d from pos_embed.py
            # head_dim is embed_dim // num_heads
            head_dim = embed_dim // num_heads
            rope_tensor = get_rope_tensor_2d(head_dim, grid_h, grid_w)
            self.register_buffer('rope_tensor', rope_tensor)
            
            # Store grid dimensions for dynamic resizing
            self.grid_h = grid_h
            self.grid_w = grid_w
           
        

        # Transformer blocks
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]
        if use_movq:
            self.blocks = nn.ModuleList([
                MoVQBlockv2(dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                           qk_norm=qk_norm, drop=proj_drop, attn_drop=attn_drop, drop_path=dpr[i],
                           latent_dim=codebook_embed_dim)
                for i in range(depth)
            ])
            self.norm = MoVQNorm(codebook_embed_dim, embed_dim)
            
            # Zero-out adaLN modulation layers
            for block in self.blocks:
                nn.init.constant_(block.adaLN_modulation[-1].weight, 0)
                nn.init.constant_(block.adaLN_modulation[-1].bias, 0)
            
            # Zero-out output layers
            nn.init.constant_(self.norm.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(self.norm.adaLN_modulation[-1].bias, 0)
        else:
            self.blocks = nn.ModuleList([
                Block(dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                      qk_norm=qk_norm, drop=proj_drop, attn_drop=attn_drop, drop_path=dpr[i])
                for i in range(depth)
            ])
            self.norm = nn.LayerNorm(embed_dim)

    def no_weight_decay(self):
        return {'pos_embed', 'cls_token', 'mask_token', 'latent_pos_embed'}

    def get_last_layer(self):
        return self.to_pixel.get_last_layer()

    def forward(self, z, interpolate_zq=None, H=None, W=None, D=None):
        B = z.shape[0]

        num_img_tokens = self.num_img_tokens

        # Create mask tokens
        if self.num_latent_tokens:
            x = self.mask_token.expand(B, num_img_tokens, -1)
        else:
            x = z

        # Add position embedding
        if self.use_ape and self.pos_embed is not None:
            x = x + self.pos_embed[:, :num_img_tokens]
        x = self.pos_drop(x)

        if self.num_latent_tokens:
            # Add latent tokens with position embedding
            z = z + self.latent_pos_embed
            x = torch.cat([x, z], dim=1)

        if self.cls_token is not None:
            x = torch.cat([self.cls_token.expand(B, -1, -1), x], dim=1)

        # Forward through transformer blocks
        if self.use_ape:
            for i, blk in enumerate(self.blocks):
                if self.use_movq:
                    x = blk(x, interpolate_zq, num_prefix_tokens=self.num_prefix_tokens, 
                           num_latent_tokens=self.num_latent_tokens)
                else:
                    x = blk(x)
        elif self.use_rope:
            # Get or compute RoPE tensor for current spatial dimensions
            if H is not None and W is not None:
                current_grid_h = H // (self.patch_size[0] if isinstance(self.patch_size, tuple) else self.patch_size)
                current_grid_w = W // (self.patch_size[1] if isinstance(self.patch_size, tuple) else self.patch_size)
            else:
                # Use default grid dimensions
                current_grid_h = self.grid_h
                current_grid_w = self.grid_w
            
            # Recompute RoPE tensor if spatial dimensions changed
            if hasattr(self, 'grid_h') and (current_grid_h != self.grid_h or current_grid_w != self.grid_w):
                head_dim = self.embed_dim // self.blocks[0].attn.num_heads
                rope_tensor = get_rope_tensor_2d(head_dim, current_grid_h, current_grid_w)
                rope_tensor = rope_tensor.to(x.device)
            else:
                rope_tensor = self.rope_tensor
            
            for i, blk in enumerate(self.blocks):
                if self.use_movq:
                    x = blk(x, interpolate_zq, rope_tensor=rope_tensor, 
                           num_prefix_tokens=self.num_prefix_tokens, num_latent_tokens=self.num_latent_tokens)
                else:
                    x = blk(x, rope_tensor=rope_tensor, num_prefix_tokens=self.num_prefix_tokens, 
                           num_latent_tokens=self.num_latent_tokens)
        else:
            for i, blk in enumerate(self.blocks):
                if self.use_movq:
                    x = blk(x, interpolate_zq, num_prefix_tokens=self.num_prefix_tokens, 
                           num_latent_tokens=self.num_latent_tokens)
                else:
                    x = blk(x)

        # Final normalization
        if self.use_movq:
            x = self.norm(x, interpolate_zq, num_prefix_tokens=self.num_prefix_tokens, 
                         num_latent_tokens=self.num_latent_tokens)
        else:
            x = self.norm(x)

        # Extract image tokens
        x = x[:, self.num_prefix_tokens:self.num_img_tokens + self.num_prefix_tokens]

        # Convert to pixel space
        out = self.to_pixel(x)
        return out


if __name__ == '__main__':
    data = torch.randn(1, 3, 224, 384)
    encoder = MAETokViTEncoder(in_channels=3, num_latent_tokens=32, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4.,
                 img_size=[224, 384], patch_size=[16, 16], drop_path_rate=0.0,
                 qkv_bias=True, qk_norm=False, attn_drop=0., proj_drop=0.,
                 rope_theta=100.0, rope_mixed=False, use_rope=False, use_ape=True,
                 base_img_size=[224, 384])
    out = encoder(data)
    ### Usually here we have the quant layers to align like in MaeTok which go enc_embed_dim -> codebook_embed_dim -> decoder_embed_dim
    decoder = MAETokViTDecoder(in_channels=3, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4.,
                 img_size=[224, 384], patch_size=[16, 16], drop_path_rate=0.0,
                 qkv_bias=True, qk_norm=False, attn_drop=0., proj_drop=0.,
                 num_latent_tokens=32, to_pixel='linear', codebook_embed_dim=32,
                 rope_theta=100.0, rope_mixed=False, use_rope=False, use_ape=True,
                 cls_token=True, base_img_size=[224, 384], use_movq=False)
    out = decoder(out)


    data = torch.randn(1, 3, 224, 224, 384)
    encoder = MAETokViTEncoder(in_channels=3, num_latent_tokens=32, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4.,
                 img_size=[224, 224, 384], patch_size=[16, 16, 16], drop_path_rate=0.0,
                 qkv_bias=True, qk_norm=False, attn_drop=0., proj_drop=0.,
                 rope_theta=100.0, rope_mixed=False, use_rope=False, use_ape=True,
                 base_img_size=[224, 224, 384])
    out = encoder(data)
    ### Usually here we have the "quant" layers to align like in MaeTok which go enc_embed_dim -> codebook_embed_dim -> decoder_embed_dim
    decoder = MAETokViTDecoder(in_channels=3, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4.,
                 img_size=[224, 224, 384], patch_size=[16, 16, 16], drop_path_rate=0.0,
                 qkv_bias=True, qk_norm=False, attn_drop=0., proj_drop=0.,
                 num_latent_tokens=32, to_pixel='linear', codebook_embed_dim=32,
                 rope_theta=100.0, rope_mixed=False, use_rope=False, use_ape=True,
                 cls_token=True, base_img_size=[224, 224, 384], use_movq=False)
    out = decoder(out)


import torch 
import torch.nn as nn
from medlat.registry import register_model
from .models import DiT

@register_model("dit.xl_1")
def DiT_XL_1(**kwargs):
    return DiT(depth=28, hidden_size=1152, patch_size=1, num_heads=16, **kwargs)

@register_model("dit.xl_2")
def DiT_XL_2(**kwargs):
    return DiT(depth=28, hidden_size=1152, patch_size=2, num_heads=16, **kwargs)

@register_model("dit.xl_4")
def DiT_XL_4(**kwargs):
    return DiT(depth=28, hidden_size=1152, patch_size=4, num_heads=16, **kwargs)

@register_model("dit.xl_8")
def DiT_XL_8(**kwargs):
    return DiT(depth=28, hidden_size=1152, patch_size=8, num_heads=16, **kwargs)

@register_model("dit.l_1")
def DiT_L_1(**kwargs):
    return DiT(depth=24, hidden_size=1024, patch_size=1, num_heads=16, **kwargs)

@register_model("dit.l_2")
def DiT_L_2(**kwargs):
    return DiT(depth=24, hidden_size=1024, patch_size=2, num_heads=16, **kwargs)

@register_model("dit.l_4")
def DiT_L_4(**kwargs):
    return DiT(depth=24, hidden_size=1024, patch_size=4, num_heads=16, **kwargs)

@register_model("dit.l_8")
def DiT_L_8(**kwargs):
    return DiT(depth=24, hidden_size=1024, patch_size=8, num_heads=16, **kwargs)

@register_model("dit.b_1")
def DiT_B_1(**kwargs):
    return DiT(depth=12, hidden_size=768, patch_size=1, num_heads=12, **kwargs)

@register_model("dit.b_2")
def DiT_B_2(**kwargs):
    return DiT(depth=12, hidden_size=768, patch_size=2, num_heads=12, **kwargs)

@register_model("dit.b_4")
def DiT_B_4(**kwargs):
    return DiT(depth=12, hidden_size=768, patch_size=4, num_heads=12, **kwargs)

@register_model("dit.b_8")
def DiT_B_8(**kwargs):
    return DiT(depth=12, hidden_size=768, patch_size=8, num_heads=12, **kwargs)

@register_model("dit.s_1")
def DiT_S_1(**kwargs):
    return DiT(depth=12, hidden_size=384, patch_size=1, num_heads=6, **kwargs)

@register_model("dit.s_2")
def DiT_S_2(**kwargs):
    return DiT(depth=12, hidden_size=384, patch_size=2, num_heads=6, **kwargs)

@register_model("dit.s_4")
def DiT_S_4(**kwargs):
    return DiT(depth=12, hidden_size=384, patch_size=4, num_heads=6, **kwargs)

@register_model("dit.s_8")
def DiT_S_8(**kwargs):
    return DiT(depth=12, hidden_size=384, patch_size=8, num_heads=6, **kwargs)


# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# timm: https://github.com/rwightman/pytorch-image-models/tree/master/timm
# DeiT: https://github.com/facebookresearch/deit
# --------------------------------------------------------

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F

from timm.models.vision_transformer import PatchEmbed, Mlp, DropPath
from typing import Dict, Optional, Tuple, Union
import os, sys

from medlat.modules.pos_embed import get_2d_sincos_pos_embed

############################## Docking Functions ##############################
from dataclasses import dataclass
from typing import Optional, Tuple
from diffusers.utils import BaseOutput
from einops import rearrange
from medlat.first_stage.modules.gaussian_dist import DiagonalGaussianDistribution
from PIL import Image
from torchvision import transforms
import numpy as np


class Config:
    def __init__(self, scaling_factor):
        self.scaling_factor = scaling_factor

@dataclass
class DecoderOutput(BaseOutput):
    r"""
    Output of decoding method.

    Args:
        sample (`torch.Tensor` of shape `(batch_size, num_channels, height, width)`):
            The decoded output sample from the last layer of the model.
    """

    sample: torch.Tensor
    commit_loss: Optional[torch.FloatTensor] = None

@dataclass
class EncoderOutput(BaseOutput):
    r"""
    Output of decoding method.

    Args:
        sample (`torch.Tensor` of shape `(batch_size, num_channels, height, width)`):
            The decoded output sample from the last layer of the model.
    """
    
    latent: torch.Tensor

    def sample(self):
        return self.latent
    def mode(self):
        return self.latent

@dataclass
class MAEOutput(BaseOutput):
    """
    Output of AutoencoderKL encoding method.

    Args:
        latent_dist (`DiagonalGaussianDistribution`):
            Encoded outputs of `Encoder` represented as the mean and logvar of `DiagonalGaussianDistribution`.
            `DiagonalGaussianDistribution` allows for sampling latents from the distribution.
    """

    latent_dist: Union[DiagonalGaussianDistribution, EncoderOutput]  # noqa: F821

def center_crop_arr(pil_image, image_size):
    """
    Center cropping implementation from ADM.
    https://github.com/openai/guided-diffusion/blob/8fb3ad9197f16bbc40620447b2742e13458d2831/guided_diffusion/image_datasets.py#L126
    """
    while min(*pil_image.size) >= 2 * image_size:
        pil_image = pil_image.resize(
            tuple(x // 2 for x in pil_image.size), resample=Image.BOX
        )

    scale = image_size / min(*pil_image.size)
    pil_image = pil_image.resize(
        tuple(round(x * scale) for x in pil_image.size), resample=Image.BICUBIC
    )

    arr = np.array(pil_image)
    crop_y = (arr.shape[0] - image_size) // 2
    crop_x = (arr.shape[1] - image_size) // 2
    return Image.fromarray(arr[crop_y: crop_y + image_size, crop_x: crop_x + image_size])

####################################################################################


class LayerScale(nn.Module):
    def __init__(self, dim, init_values=1e-5, inplace=False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x):
        return x.mul_(self.gamma) if self.inplace else x * self.gamma
    
class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0., proj_drop=0.):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, return_attn_map = False):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        if return_attn_map:
            qk_attn = attn.clone().detach()
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x_ctxed = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x_ctxed)
        x = self.proj_drop(x)
        
        if return_attn_map:
            return x, [qk_attn, x_ctxed]
        return x
    
class Block(nn.Module):

    def __init__(
            self,
            dim,
            num_heads,
            mlp_ratio=4.,
            qkv_bias=False,
            drop=0.,
            attn_drop=0.,
            init_values=None,
            drop_path=0.,
            act_layer=nn.GELU,
            norm_layer=nn.LayerNorm
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)
        self.ls1 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim)
        self.mlp = Mlp(in_features=dim, hidden_features=int(dim * mlp_ratio), act_layer=act_layer, drop=drop)
        self.ls2 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x, return_attn_map = False):
        if return_attn_map:
            x_tmp, qk_and_x = self.attn(self.norm1(x), return_attn_map = True)
            # returned attn is Q @ K / scale, before softmax
        else:
            x_tmp = self.attn(self.norm1(x))
        x = x + self.drop_path1(self.ls1(x_tmp))
        x = x + self.drop_path2(self.ls2(self.mlp(self.norm2(x))))
        
        if return_attn_map:
            return x, qk_and_x
        return x

class Downsample(nn.Module): # from downblock in AutoencoderKL
    def __init__(self, ch_in, ch_out):
        super().__init__()
        self.conv = nn.Conv2d(ch_in, ch_out, 3, stride=2)
        
    def forward(self, x): # x : B N C
        B, N, C = x.shape
        H = int(N ** 0.5)
        assert H * H == N, 'Size mismatch.'
        x = x.reshape(B, H, H, C).permute(0,3,1,2) # B C H W
        
        pad = (0, 1, 0, 1)
        x = torch.nn.functional.pad(x, pad, mode="constant", value=0)
        x = self.conv(x) # B C H W
        
        x = x.reshape(B, C, -1).permute(0,2,1) # B N C
        return x
    
class Upsample(nn.Module): # from upblock in AutoencoderKL
    def __init__(self, ch_in, ch_out):
        super().__init__()
        self.conv = nn.Conv2d(ch_in, ch_out, 3, 1, 1)
        
    def forward(self, x): # x : B N C
        B, N, C = x.shape
        H = int(N ** 0.5)
        assert H * H == N, 'Size mismatch.'
        x = x.reshape(B, H, H, C).permute(0,3,1,2) # B C H W
        
        if x.shape[0] >= 64:
            x = x.contiguous()
            
        scale_factor = 2
        if x.numel() * scale_factor > pow(2, 31):
            x = x.contiguous()

        x = F.interpolate(x, scale_factor=2.0, mode="nearest")
        
        x = self.conv(x)
        
        x = x.reshape(B, C, -1).permute(0,2,1) # B N C
        return x
    
class MLP_dim_resize(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim):
        super(MLP_dim_resize, self).__init__()
        self.layers = nn.Sequential(
            nn.Linear(input_dim, hidden_dim), 
            nn.GELU(),                         
            nn.Linear(hidden_dim, output_dim) 
        )

    def forward(self, x):
        return self.layers(x)
    
class conv_decoder_pred(nn.Module):
    def __init__(self, decoder_embed_dim, patch_size, in_chans, pred_with_conv=True):
        super(conv_decoder_pred, self).__init__()
        self.p = patch_size
        self.pred_with_conv = pred_with_conv
        if self.pred_with_conv:
            print('pred only with conv instead of previous linear')
            self.conv_smoother = nn.Conv2d(decoder_embed_dim, patch_size**2 * in_chans, 2, stride=1, padding=0)
        else:
            print('conv on rgb')
            self.linear_pred = nn.Linear(decoder_embed_dim, patch_size**2 * in_chans, bias=True)
            self.conv_smoother = nn.Conv2d(in_chans, in_chans, 3, 1, 1)
            
    def forward(self, x):
        h = w = int(x.shape[1]**.5)
        assert h * w == x.shape[1]
        
        if self.pred_with_conv:
            B = x.shape[0]
            x = x.reshape(B, h, w, -1).permute(0,3,1,2)
            padding = (0, 1, 0, 1)  # Pad 1 on the right (W) and 1 on the bottom (H)
            # Apply padding
            x = F.pad(x, padding, mode='constant', value=0)
            x  = self.conv_smoother(x) # B C H W
            x = x.reshape(B, -1, h*w).permute(0,2,1) # B HW C
            
        else:
            x = self.linear_pred(x) # B HW p_size*p_size*3
            x = x.reshape(shape=(x.shape[0], h, w, self.p, self.p, 3))
            x = torch.einsum('nhwpqc->nchpwq', x)
            x = x.reshape(shape=(x.shape[0], 3, h * self.p, w * self.p)) # B 3 256 256
            
            x = self.conv_smoother(x)
            x = x.reshape(x.shape[0], 3, h, self.p, w, self.p)
            x = torch.einsum('nchpwq->nhwpqc', x)
            x = x.reshape(shape=(x.shape[0], h*w, self.p*self.p*3)) # B HW C
        
        return x

class MaskedAutoencoderViT(nn.Module):
    """ Masked Autoencoder with VisionTransformer backbone
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3,
                 embed_dim=1024, depth=24, num_heads=16,
                 decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
                 mlp_ratio=4., norm_layer=nn.LayerNorm, norm_pix_loss=False,
                 latent_dim=32, ldmae_mode=False, scaling_factor=0.9654248952865601, no_cls=True, 
                 gradual_resol=False, finetune_downsample_layer=None, down_nonlinear=False,
                 kl_loss_weight=None, smooth_output=False, pred_with_conv=False, perceptual_loss=None, perceptual_loss_ratio=1.0,
                 fixed_std=None):
        super().__init__()
        
        # --------------------------------------------------------------------------
        # MAE for LDMAE settings
        self.fixed_std = fixed_std
        self.perceptual_loss = perceptual_loss
        self.perceptual_loss_ratio = perceptual_loss_ratio
        self.smooth_output = smooth_output
        self.gradual_resol = gradual_resol
        self.kl_loss_weight = kl_loss_weight
        encoder_latent_dim = latent_dim
        decoder_latent_dim = latent_dim
        if self.kl_loss_weight is not None: 
            assert no_cls, 'There should be no class token to use KL loss.'
            encoder_latent_dim = 2 * latent_dim
            print(f'Use KL loss, encoder latent dim is {encoder_latent_dim} to predict mean & logvar')
        if self.gradual_resol:
            patch_size = patch_size // 2
            print(f'patch size: {patch_size}')
        
        if down_nonlinear:
            print('Use MLP for latent embedding')
            self.to_latent = MLP_dim_resize(embed_dim, latent_dim*4, encoder_latent_dim)
            self.from_latent = MLP_dim_resize(decoder_latent_dim, latent_dim*4, embed_dim)
        else:
            self.to_latent = nn.Linear(embed_dim, encoder_latent_dim)
            self.from_latent = nn.Linear(decoder_latent_dim, embed_dim)
        self.config = Config(scaling_factor=scaling_factor)
        self.ldmae_mode = ldmae_mode
        self.img_size = img_size
        self.patch_size = patch_size
        self.latent_resolution = img_size // patch_size
        self.tile_latent_min_size = self.latent_resolution
        self.latent_dim = latent_dim
        self.no_cls = no_cls
        self.num_extra_tokens = 0
        # if head_output_num is not None:
        #     self.head = torch.nn.ModuleList([
        #         torch.nn.LayerNorm(self.latent_dim),
        #         # torch.nn.BatchNorm1d(self.latent_dim, affine=False, eps=1e-6),
        #         torch.nn.Linear(in_features=self.latent_dim, out_features=head_output_num, bias=False)
        #     ])
            # self.head_norm = torch.nn.LayerNorm(self.latent_dim)
            # self.head_bn = torch.nn.BatchNorm1d(self.latent_dim, affine=False, eps=1e-6)
            # self.head = torch.nn.Linear(in_features=self.latent_dim, out_features=head_output_num)
        if not self.no_cls:
            self.num_extra_tokens += 1
        # --------------------------------------------------------------------------
        
        # --------------------------------------------------------------------------
        # MAE encoder specifics
        self.patch_embed = PatchEmbed(img_size, patch_size, in_chans, embed_dim)
        num_patches = self.patch_embed.num_patches

        if not self.no_cls:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + self.num_extra_tokens, embed_dim), requires_grad=False)  # fixed sin-cos embedding
        
        if self.gradual_resol:
            blocks = []
            downsize_time = depth // 2 if finetune_downsample_layer is None else finetune_downsample_layer
            for i in range(depth):
                blocks.append(Block(embed_dim, num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer))
                if i == downsize_time-1:
                    print(f"Add downsizing block in {i}th layer in encoder.")
                    blocks.append(Downsample(embed_dim, embed_dim))
            self.blocks = nn.ModuleList(blocks)
        else:
            self.blocks = nn.ModuleList([
                Block(embed_dim, num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer)
            for i in range(depth)])
        
        self.norm = norm_layer(embed_dim)
        # --------------------------------------------------------------------------

        # --------------------------------------------------------------------------
        # MAE decoder specifics
        self.decoder_embed = nn.Linear(embed_dim, decoder_embed_dim, bias=True)

        if not self.ldmae_mode:
            self.mask_token = nn.Parameter(torch.zeros(1, 1, decoder_embed_dim))

        if self.gradual_resol:
            decoder_num_patches = num_patches // 4
            assert decoder_num_patches * 4 == num_patches
        else:
            decoder_num_patches = num_patches
        self.decoder_pos_embed = nn.Parameter(torch.zeros(1, decoder_num_patches + self.num_extra_tokens, decoder_embed_dim), requires_grad=False)  # fixed sin-cos embedding

        if self.gradual_resol:
            decoder_blocks = []
            upsize_time = decoder_depth - downsize_time
            for i in range(decoder_depth):
                decoder_blocks.append(Block(decoder_embed_dim, decoder_num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer))
                if i == upsize_time-1:
                    print(f"Add upsizing block in {i}th layer in decoder.")
                    decoder_blocks.append(Upsample(decoder_embed_dim, decoder_embed_dim))
            self.decoder_blocks = nn.ModuleList(decoder_blocks)
        else:
            self.decoder_blocks = nn.ModuleList([
                Block(decoder_embed_dim, decoder_num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer)
                for i in range(decoder_depth)])
            
        self.decoder_norm = norm_layer(decoder_embed_dim)
        if smooth_output:
            assert no_cls, 'Should be no CLS token for smooth_output.'
            print('Use conv in decoder pred.')
            self.decoder_pred = conv_decoder_pred(decoder_embed_dim, patch_size, in_chans, pred_with_conv)
        else:
            self.decoder_pred = nn.Linear(decoder_embed_dim, patch_size**2 * in_chans, bias=True) # decoder to patch
        # --------------------------------------------------------------------------

        self.norm_pix_loss = norm_pix_loss

        self.initialize_weights()
    
    
    def initialize_weights(self):
        # initialization
        # initialize (and freeze) pos_embed by sin-cos embedding
        pos_embed = get_2d_sincos_pos_embed(self.pos_embed.shape[-1], int(self.patch_embed.num_patches**.5), cls_token= not self.no_cls)
        self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

        if self.gradual_resol:
            decoder_num_patches = self.patch_embed.num_patches // 4
        else:
            decoder_num_patches = self.patch_embed.num_patches
        decoder_pos_embed = get_2d_sincos_pos_embed(self.decoder_pos_embed.shape[-1], int(decoder_num_patches**.5), cls_token= not self.no_cls)
        self.decoder_pos_embed.data.copy_(torch.from_numpy(decoder_pos_embed).float().unsqueeze(0))

        # initialize patch_embed like nn.Linear (instead of nn.Conv2d)
        w = self.patch_embed.proj.weight.data
        torch.nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

        # timm's trunc_normal_(std=.02) is effectively normal_(std=0.02) as cutoff is too big (2.)
        if not self.no_cls:
            torch.nn.init.normal_(self.cls_token, std=.02)
        if not self.ldmae_mode:
            torch.nn.init.normal_(self.mask_token, std=.02)

        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def patchify(self, imgs):
        """
        imgs: (N, 3, H, W)
        x: (N, L, patch_size**2 *3)
        """
        p = self.patch_embed.patch_size[0]
        assert imgs.shape[2] == imgs.shape[3] and imgs.shape[2] % p == 0

        h = w = imgs.shape[2] // p
        x = imgs.reshape(shape=(imgs.shape[0], 3, h, p, w, p))
        x = torch.einsum('nchpwq->nhwpqc', x)
        x = x.reshape(shape=(imgs.shape[0], h * w, p**2 * 3))
        return x

    def unpatchify(self, x):
        """
        x: (N, L, patch_size**2 *3)
        imgs: (N, 3, H, W)
        """
        p = self.patch_embed.patch_size[0]
        h = w = int(x.shape[1]**.5)
        assert h * w == x.shape[1]
        
        x = x.reshape(shape=(x.shape[0], h, w, p, p, 3))
        x = torch.einsum('nhwpqc->nchpwq', x)
        imgs = x.reshape(shape=(x.shape[0], 3, h * p, h * p))
        return imgs

    def random_masking(self, x, mask_ratio):
        """
        Perform per-sample random masking by per-sample shuffling.
        Per-sample shuffling is done by argsort random noise.
        x: [N, L, D], sequence
        """
        N, L, D = x.shape  # batch, length, dim
        len_keep = int(L * (1 - mask_ratio))
        
        noise = torch.rand(N, L, device=x.device)  # noise in [0, 1]
        
        # sort noise for each sample
        ids_shuffle = torch.argsort(noise, dim=1)  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        x_masked = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return x_masked, mask, ids_restore

    def forward_encoder(self, x, mask_ratio):
        # embed patches
        x = self.patch_embed(x)
        
        # add pos embed w/o cls token
        if self.no_cls:
            x = x + self.pos_embed
        else:
            x = x + self.pos_embed[:, 1:, :]

        # masking: length -> length * mask_ratio
        x, mask, ids_restore = self.random_masking(x, mask_ratio)

        # append cls token
        if not self.no_cls:
            cls_token = self.cls_token + self.pos_embed[:, :1, :]
            cls_tokens = cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)

        return x, mask, ids_restore

    def forward_decoder(self, x, ids_restore):
        # embed tokens
        x = self.decoder_embed(x)

        # append mask tokens to sequence
        mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        if not self.no_cls:
            x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
            x_ = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle
            x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token
        else:
            x_ = torch.cat([x, mask_tokens], dim=1)  # no cls token
            x = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle

        # add pos embed
        x = x + self.decoder_pos_embed

        # apply Transformer blocks
        for blk in self.decoder_blocks:
            x = blk(x)
        x = self.decoder_norm(x)

        # predictor projection
        x = self.decoder_pred(x)

        # remove cls token
        if not self.no_cls:
            x = x[:, 1:, :]

        return x
    
    def forward_encoder_with_mask(self, x, mask_ratio):
        # embed patches
        x = self.patch_embed(x)
        
        # add pos embed w/o cls token
        # if self.no_cls:
        #     x = x + self.pos_embed
        # else:
        #     x = x + self.pos_embed[:, 1:, :]

        # masking: length -> length * mask_ratio
        x, mask, ids_restore = self.random_masking(x, mask_ratio)
        
        # append mask tokens to sequence
        mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        # if not self.no_cls:
        #     x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
        #     x_ = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle
        #     x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token
        # else:
        x_ = torch.cat([x, mask_tokens], dim=1) 
        x = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle

        # append cls token
        if self.no_cls:
            x = x + self.pos_embed
        else:
            x = x + self.pos_embed[:, 1:, :]
            cls_token = self.cls_token + self.pos_embed[:, :1, :]
            cls_tokens = cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)

        return x, mask, ids_restore

    def forward_decoder_without_mask(self, x, ids_restore):
        # embed tokens
        x = self.decoder_embed(x)

        # append mask tokens to sequence
        # mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        # if not self.no_cls:
        #     x_ = torch.cat([x[:, 1:, :], mask_tokens], dim=1)  # no cls token
        #     x_ = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle
        #     x = torch.cat([x[:, :1, :], x_], dim=1)  # append cls token
        # else:
        #     x_ = torch.cat([x, mask_tokens], dim=1)  # no cls token
        #     x = torch.gather(x_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))  # unshuffle

        # add pos embed
        x = x + self.decoder_pos_embed

        # apply Transformer blocks
        for blk in self.decoder_blocks:
            x = blk(x)
        x = self.decoder_norm(x)

        # predictor projection
        x = self.decoder_pred(x)

        # remove cls token
        if not self.no_cls:
            x = x[:, 1:, :]

        return x
    
    # self.no_cls에 따라 다르게 encoding & decoding
    # decoding은 이후에 unpatchfy 넣어주어야함
    def ldmae_encoding(self, imgs, use_mode=False, return_kl=False):
        # encoding
        x = self.patch_embed(imgs)
        if not self.no_cls:
            x = x + self.pos_embed[:, 1:, :]
        else:
            x = x + self.pos_embed
        if not self.no_cls:
            cls_token = self.cls_token + self.pos_embed[:, :1, :]
            cls_tokens = cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)
        # to latent
        latent = self.to_latent(x)

        if self.kl_loss_weight is not None:
            latent = latent.permute(0,2,1) # B D HW
            posterior = DiagonalGaussianDistribution(latent) # B D HW
            if use_mode:
                print('mode')
                latent = posterior.mode() # B D HW
            else:
                latent = posterior.sample() # B D HW
            latent = latent.permute(0,2,1)
        if return_kl:
            return latent, posterior.kl()
        return latent
    
    def ldmae_decoding(self, x):
        # if self.smooth_output:
        #     with torch.no_grad():
        #         # from latent
        #         x = self.from_latent(x)
                
        #         x = self.decoder_embed(x)
        #         # if x.shape[1] != self.decoder_pos_embed.shape[1]:
        #         if not self.no_cls:
        #             decoder_pos_embed = self.decoder_pos_embed[:,1:,:]
        #         else:
        #             decoder_pos_embed = self.decoder_pos_embed
        #         x = x + decoder_pos_embed
        #         for blk in self.decoder_blocks:
        #             x = blk(x)
        #         x = self.decoder_norm(x)
        # else:
        # from latent
        x = self.from_latent(x)
        
        x = self.decoder_embed(x)
        # if x.shape[1] != self.decoder_pos_embed.shape[1]:
        if not self.no_cls:
            decoder_pos_embed = self.decoder_pos_embed[:,1:,:]
        else:
            decoder_pos_embed = self.decoder_pos_embed
        x = x + decoder_pos_embed
        for blk in self.decoder_blocks:
            x = blk(x)
        x = self.decoder_norm(x)
            
        x = self.decoder_pred(x)
        if not self.no_cls:
            x = x[:, 1:, :]
        return x
    
    def reconstruct(self, imgs, use_mode=True, return_kl=False, mask_ratio=0.75):
        # return it back
        if mask_ratio == 0.0:
            x = self.ldmae_encoding(imgs, use_mode=True, return_kl=False)
        else:
            with torch.no_grad():
                if return_kl:
                    x, kl_val = self.ldmae_encoding(imgs, use_mode=use_mode, return_kl=return_kl)
                else:
                    x = self.ldmae_encoding(imgs, use_mode=use_mode, return_kl=return_kl)
        x = self.ldmae_decoding(x)
        if return_kl:
            return x, kl_val
        return x
    
    def linear_probe_seg(self, images):
        with torch.no_grad():
            x = self.ldmae_encoding(images)

        if not self.no_cls:
            x = x[:, 1:, :] # B HW D
        
        B, N, D = x.shape
        x = x.reshape(-1, D) # BHW D
        
        for layer in self.head:
            x = layer(x)
        # x = self.head(x)
        return x # BHW num_classes
    
    def linear_probe(self, images):
        with torch.no_grad():
            x = self.ldmae_encoding(images)
            
        # x = self.head_norm(x)
        if self.no_cls:
            x = x.mean(dim=1)  # global pool, [B D]
        else:
            x = x[:, 1:, :].mean(dim=1)  # global pool, [B D]
        # x = self.head_bn(x)
        for layer in self.head:
            x = layer(x)
        return x
    
    def forward_loss(self, imgs, pred, mask, visible_loss_ratio=0.5):
        """
        imgs: [N, 3, H, W]
        pred: [N, L, p*p*3]
        mask: [N, L], 0 is keep, 1 is remove, 
        """
        target = self.patchify(imgs)
        if self.norm_pix_loss:
            mean = target.mean(dim=-1, keepdim=True)
            var = target.var(dim=-1, keepdim=True)
            target = (target - mean) / (var + 1.e-6)**.5

        loss = (pred - target) ** 2
        loss = loss.mean(dim=-1)  # [B, L], mean loss per patch
        
        visible_loss = (loss * (1-mask)).sum() / (1-mask).sum()
        mask_loss = (loss * mask).sum() / mask.sum() # now per pixel, scalar
        
        loss = (1-visible_loss_ratio) * mask_loss + visible_loss_ratio * visible_loss
        
        if self.perceptual_loss is not None:
            p_loss = self.perceptual_loss(
                imgs.contiguous(),
                self.unpatchify(pred).contiguous()
            ) # B 1 1 1
            p_loss = p_loss.mean()
            loss = loss + self.perceptual_loss_ratio * p_loss
        else:
            p_loss = torch.zeros_like(loss)
        # loss = (loss * mask).sum() / mask.sum()  # mean loss on removed patches
        return loss, visible_loss, mask_loss, p_loss
    
    def forward_vanilla(self, imgs, mask_ratio=0.75, visible_loss_ratio=0.5):
        if self.gradual_resol:
            latent, mask, ids_restore = self.forward_encoder_with_mask(imgs, mask_ratio)
        else:
            latent, mask, ids_restore = self.forward_encoder(imgs, mask_ratio)
            
        # --------------------------------------------------------------------------
        # to latent
        latent = self.to_latent(latent)

        if self.kl_loss_weight is not None:
            B, N, D = latent.shape
            latent = latent.permute(0,2,1) # B D HW
            posterior = DiagonalGaussianDistribution(latent, fixed_std=self.fixed_std) # B D HW
            
            kl_loss = posterior.kl()
            kl_loss = torch.sum(kl_loss) / kl_loss.shape[0] / N # per patch
            
            latent = posterior.sample() # B D HW
            latent = latent.permute(0,2,1)
            
        latent = self.from_latent(latent)
        # --------------------------------------------------------------------------
        
        if self.gradual_resol:
            pred = self.forward_decoder_without_mask(latent, ids_restore)  # [N, L, p*p*3]
        else:
            pred = self.forward_decoder(latent, ids_restore)  # [N, L, p*p*3]
            
        loss, vis_loss, mask_loss, p_loss = self.forward_loss(imgs, pred, mask, visible_loss_ratio)
        if self.kl_loss_weight is not None:
            loss = loss + self.kl_loss_weight * kl_loss
        else:
            kl_loss = None
        return loss, pred, mask, vis_loss, mask_loss, kl_loss, p_loss
    
    def forward_ldmae(self, imgs, mask_ratio=0.75):
        pred = self.reconstruct(imgs, use_mode=False, mask_ratio=mask_ratio)
        
        # target = self.patchify(imgs)
        vis_loss = (self.unpatchify(pred) - imgs) ** 2
        
        if self.perceptual_loss is not None:
            p_loss = self.perceptual_loss(
                imgs.contiguous(),
                self.unpatchify(pred).contiguous()
            )
            loss = vis_loss + self.perceptual_loss_ratio * p_loss
        else:
            loss = vis_loss
            p_loss = torch.zeros_like(vis_loss)
        
        loss = loss.mean()
        return loss, pred, None, vis_loss.mean(), p_loss.mean(), None
    
    def forward(self, imgs, mask_ratio=0.75, visible_loss_ratio=0.5):
        if self.ldmae_mode:
            return self.forward_ldmae(imgs, mask_ratio=mask_ratio)
        else:
            return self.forward_vanilla(imgs, mask_ratio=mask_ratio, visible_loss_ratio=visible_loss_ratio)
    
    ############################ Docking Functions ####################################
    
    def _encode(self, x):
        # encoding
        x = self.patch_embed(x)
        if not self.no_cls:
            x = x + self.pos_embed[:, 1:, :]
        else:
            x = x + self.pos_embed
        if not self.no_cls:
            cls_token = self.cls_token + self.pos_embed[:, :1, :]
            cls_tokens = cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)
        # to latent
        x = self.to_latent(x)
        x = rearrange(x, 'b (h w) c -> b c h w', h=self.latent_resolution, w=self.latent_resolution)
        return x
    
    def encode_gflops(self, x, return_dict=True):
        # encoding
        x = self.patch_embed(x)
        if not self.no_cls:
            x = x + self.pos_embed[:, 1:, :]
        else:
            x = x + self.pos_embed
        if not self.no_cls:
            cls_token = self.cls_token + self.pos_embed[:, :1, :]
            cls_tokens = cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)
        # to latent
        x = self.to_latent(x)
        x = rearrange(x, 'b (h w) c -> b c h w', h=self.latent_resolution, w=self.latent_resolution)
        
        return x
    
    def encode(self, x, return_dict=True):
        # encoding
        x = self.patch_embed(x)
        if not self.no_cls:
            x = x + self.pos_embed[:, 1:, :]
        else:
            x = x + self.pos_embed
        if not self.no_cls:
            cls_token = self.cls_token + self.pos_embed[:, :1, :]
            cls_tokens = cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)
        # to latent
        x = self.to_latent(x)
        x = rearrange(x, 'b (h w) c -> b c h w', h=self.latent_resolution, w=self.latent_resolution)

        if self.kl_loss_weight is not None:
            p = DiagonalGaussianDistribution(x) # B c h w
        else:
            p = EncoderOutput(x)
        if not return_dict:
            return (p,)
        
        return MAEOutput(latent_dist=p)

    def decode(self, z, return_dict=True, generator=None):
        
        # from latent
        z = rearrange(z, 'b c h w -> b (h w) c')
        x = self.from_latent(z)
        
        x = self.decoder_embed(x)
        # if x.shape[1] != self.decoder_pos_embed.shape[1]:
        if not self.no_cls:
            decoder_pos_embed = self.decoder_pos_embed[:,1:,:]
        else:
            decoder_pos_embed = self.decoder_pos_embed
        x = x + decoder_pos_embed
        for blk in self.decoder_blocks:
            x = blk(x)
        x = self.decoder_norm(x)
        x = self.decoder_pred(x)

        img = self.unpatchify(x)
        if not return_dict:
            return (img,)
        
        return DecoderOutput(sample=img)

    @property
    def device(self) -> torch.device:
        """
        Returns:
            torch.device: The torch device on which the model's parameters are located.
        """
        # Check if the model has parameters that specify a device
        for param in self.parameters():
            return param.device

        # Check if the model has buffers (like self.pos_embed or self.cls_token)
        for buffer in self.buffers():
            return buffer.device

        # Fallback to CPU if no parameters or buffers are found
        return torch.device("cpu")
    
    @property
    def _execution_device(self):
        r"""
        Returns the device on which the pipeline's models will be executed. After calling
        [`~DiffusionPipeline.enable_sequential_cpu_offload`] the execution device can only be inferred from
        Accelerate's module hooks.
        """
        return self.device
    
    @property
    def dtype(self) -> torch.dtype:
        """
        Returns:
            torch.dtype: The torch data type on which the model's parameters are stored.
        """
        # Check if the model has any parameters to infer dtype
        for param in self.parameters():
            return param.dtype

        # Check if the model has buffers to infer dtype
        for buffer in self.buffers():
            return buffer.dtype

        # Default to float32 if no parameters or buffers are found
        return torch.float32
    
    #################################################################################################
    
    ####################################### VAVAE Docking Functions #################################
    def img_transform(self, p_hflip=0, img_size=None):
        """Image preprocessing transforms
        Args:
            p_hflip: Probability of horizontal flip
            img_size: Target image size, use default if None
        Returns:
            transforms.Compose: Image transform pipeline
        """
        img_size = img_size if img_size is not None else self.img_size
        img_transforms = [
            transforms.Lambda(lambda pil_image: center_crop_arr(pil_image, img_size)),
            transforms.RandomHorizontalFlip(p=p_hflip),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5], inplace=True)
        ]
        return transforms.Compose(img_transforms)
    
    def encode_images(self, images):
        """Encode images to latent representations
        Args:
            images: Input image tensor
        Returns:
            torch.Tensor: Encoded latent representation
        """
        with torch.no_grad():
            posterior = self.encode(images.cuda(), return_dict=False)[0]
            return posterior.sample()

    def decode_to_images(self, z):
        """Decode latent representations to images
        Args:
            z: Latent representation tensor
        Returns:
            np.ndarray: Decoded image array
        """
        with torch.no_grad():
            images = self.decode(z.cuda(), return_dict=False)[0]
            images = torch.clamp(127.5 * images + 128.0, 0, 255).permute(0, 2, 3, 1).to("cpu", dtype=torch.uint8).numpy()
        return images
    
    ############################################################################################

def mae_for_ldmae(**kwargs):
    model = MaskedAutoencoderViT(
        img_size=128, patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=32, **kwargs)
    return model

# 위랑 같은건데 그냥 복붙해서 이름만 바꿈
def mae_for_ldmae_f8d32(**kwargs):
    model = MaskedAutoencoderViT(
        img_size=128, patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=32, **kwargs)
    return model

def mae_for_ldmae_f8d16_prev(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, **kwargs)
    return model

def mae_for_ldmae_f8d16_small(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=96, depth=12, num_heads=8,
        decoder_embed_dim=96, decoder_depth=12, decoder_num_heads=8,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, **kwargs)
    return model

def mae_for_ldmae_f8d16_asym_small(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=96, depth=12, num_heads=8,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, **kwargs)
    return model

def mae_for_ldmae_f8d16_prev_large(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=384, depth=12, num_heads=16,
        decoder_embed_dim=384, decoder_depth=12, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, **kwargs)
    return model

def mae_for_ldmae_f8d16(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=384, decoder_depth=12, decoder_num_heads=24,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, down_nonlinear=True,**kwargs)
    return model

def mae_for_ldmae_f8d16_flexible(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=384, decoder_depth=12, decoder_num_heads=24,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, down_nonlinear=True,**kwargs)
    return model

def mae_for_ldmae_f16d32(**kwargs):
    model = MaskedAutoencoderViT(
        img_size=128, patch_size=16, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=32, **kwargs)
    return model

def mae_for_ldmae_f16d32_large(**kwargs):
    model = MaskedAutoencoderViT(
        img_size=128, patch_size=16, embed_dim=384, depth=12, num_heads=12,
        decoder_embed_dim=384, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=32, finetune_downsample_layer=4, **kwargs)
    return model

def mae_for_ldmae_f8d32_flexible(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=32, **kwargs)
    return model

def mae_for_ldmae_16d(**kwargs):
    model = MaskedAutoencoderViT(
        img_size=128, patch_size=8, embed_dim=192, depth=12, num_heads=12,
        decoder_embed_dim=192, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), latent_dim=16, **kwargs)
    return model

def mae_vit_base_patch16_dec512d8b(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=16, embed_dim=768, depth=12, num_heads=12,
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model

def mae_vit_base_patch16_dec128d8b(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=16, embed_dim=768, depth=12, num_heads=12,
        decoder_embed_dim=128, decoder_depth=8, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_vit_large_patch16_dec512d8b(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=16, embed_dim=1024, depth=24, num_heads=16,
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_vit_huge_patch14_dec512d8b(**kwargs):
    model = MaskedAutoencoderViT(
        patch_size=14, embed_dim=1280, depth=32, num_heads=16,
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


# set recommended archs
mae_vit_base_patch16 = mae_vit_base_patch16_dec512d8b  # decoder: 512 dim, 8 blocks
mae_vit_large_patch16 = mae_vit_large_patch16_dec512d8b  # decoder: 512 dim, 8 blocks
mae_vit_huge_patch14 = mae_vit_huge_patch14_dec512d8b  # decoder: 512 dim, 8 blocks
mae_vit_base_patch16_128 = mae_vit_base_patch16_dec128d8b